/*
* Copyright (c) 2008, Mike Giles
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * The name of Mike Giles may not be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY Mike Giles ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Mike Giles BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


/* CPU version of generic solver for 3D PDE's */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <myvar.h>


void CN_CPU(REAL xmin[],REAL dx[],REAL dxi[],REAL dt,
            int option, int I, int J, int K,
            REAL A0[],REAL A1[],REAL A2[],REAL R[],REAL V[],REAL R2[], REAL consts[]);
void trid(REAL* a, REAL * b, REAL * c, REAL* d, int NX, int inc);
void trid1(REAL* a, REAL * b, REAL * c, REAL* d, int NX, int inc);
void adi3(int option, int J, int K, int L,
          REAL A0[], REAL A1[], REAL A2[],
          REAL V[], REAL R[], REAL R2[]);


/* top level driver for generic 3-factor PDE with abritrary drift and volatility */

void generic3_CPU(REAL Tmax, int M, REAL xmin[], REAL xmax[],int I, int J, int K,
                  REAL V[], char option[], REAL consts[], int n_consts)
{
  int    IJK, m;
  REAL   dt, dx[3], dxi[3];
  REAL   *A0, *A1, *A2, *R,*R2;

/* allocate host arrays, with padding as needed */
  
  IJK = I*J*K;

  A0  = (REAL *) calloc(3*IJK, sizeof(REAL));
  A1  = (REAL *) calloc(3*IJK, sizeof(REAL));
  A2  = (REAL *) calloc(3*IJK, sizeof(REAL));
  R   = (REAL *) calloc(  IJK+2*(I+1)*(J+1), sizeof(REAL)) + (I+1)*(J+1);
  R2  = (REAL *) calloc(  IJK+2*(I+1)*(J+1), sizeof(REAL)) + (I+1)*(J+1);

/* grid spacing */

  dx[0]  = (xmax[0]-xmin[0]) / (I-1);
  dx[1]  = (xmax[1]-xmin[1]) / (J-1);
  dx[2]  = (xmax[2]-xmin[2]) / (K-1);
    
  dxi[0] = 1.0/dx[0];
  dxi[1] = 1.0/dx[1];
  dxi[2] = 1.0/dx[2];

  dt = Tmax / M;

/* main Crank-Nicolson time-marching loop */
  
  for (m=0; m<M; m++) {

/* standard timestep */

    if (strcmp(option,"CN1")==0)
      CN_CPU(xmin,dx,dxi,dt, 0, I,J,K,A0,A1,A2,R,V,R2, consts);

/* Craig-Sneyd timestep */

    else if (strcmp(option,"CN2")==0)
      CN_CPU(xmin,dx,dxi,dt, 1, I,J,K,A0,A1,A2,R,V,R2, consts);

  }

  free(A0);
  free(A1);
  free(A2);
  free(R  - (I+1)*(J+1));
  free(R2 - (I+1)*(J+1));
}



/* CPU version of Crank-Nicolson solver called by generic 3-factor solver */

void CN_CPU(REAL xmin[],REAL dx[],REAL dxi[],REAL dt,
            int option, int I, int J, int K,
            REAL A0[],REAL A1[],REAL A2[],REAL R[],REAL V[],REAL R2[], REAL consts[])
{
  REAL x[3],u[3],src[3],sig[3],beta[3];
	
  int  ind, i,j,k, i0,i1,i2, IJK;
  REAL a0m,a0o,a0p,b0, a1m,a1o,a1p,b1, a2m,a2o,a2p,b2;

  IJK = I*J*K;

  for (k=0; k<K; k++){
    for (j=0; j<J; j++) {
      for (i=0; i<I; i++) {

        ind = i + j*I + k*I*J;

        params(i,j,k,I,J,K,dt,x,dx,xmin,dxi, src,u,sig,beta, consts); 

        a0m =      sig[0] - u[0];
        a0o = -2.0*sig[0] - src[0];
        a0p =      sig[0] + u[0];

        a1m =      sig[1] - u[1];
        a1o = -2.0*sig[1] - src[1];
        a1p =      sig[1] + u[1];

        a2m =      sig[2] - u[2];
        a2o = -2.0*sig[2] - src[2];
        a2p =      sig[2] + u[2];

        i0 = (beta[0]<0? -1:1);
        i1 = (beta[1]<0? -1:1);
        i2 = (beta[2]<0? -1:1);

        b0 = fabs(beta[0]);
        b1 = fabs(beta[1]);
        b2 = fabs(beta[2]);

        R[ind] = (a0m-b1-b2)*V[ind-1]   + (a0p-b1-b2)*V[ind+1] 
               + (a1m-b2-b0)*V[ind-I]   + (a1p-b2-b0)*V[ind+I]
               + (a2m-b0-b1)*V[ind-I*J] + (a2p-b0-b1)*V[ind+I*J]
               + (a0o+a1o+a2o+2.0*(b0+b1+b2))*V[ind]
               + b2*(V[ind-1-i2*I]   + V[ind+1+i2*I])
               + b1*(V[ind-1-i1*I*J] + V[ind+1+i1*I*J])
               + b0*(V[ind-J-i0*I*J] + V[ind+J+i0*I*J]);
	
        A0[ind      ] =     - 0.5*a0m;
        A0[ind+IJK  ] = 1.0 - 0.5*a0o;
        A0[ind+2*IJK] =     - 0.5*a0p;

        A1[ind      ] =     - 0.5*a1m;
        A1[ind+IJK  ] = 1.0 - 0.5*a1o;
        A1[ind+2*IJK] =     - 0.5*a1p;

        A2[ind      ] =     - 0.5*a2m;
        A2[ind+IJK  ] = 1.0 - 0.5*a2o;
        A2[ind+2*IJK] =     - 0.5*a2p;

      }
    }
  }

  adi3(0, I,J,K, A0,A1,A2, V,R,R2);

  // Craig-Sneyd correction, to be added later

  /*
  if (option==1) {

  }
  */
}



/*
   set of low-level routines for 3D ADI time-marching

   trid       tri-diagonal solver
   trid1      second version re-using factored matrices

   adi3       main 3D ADI time-marching routine

*/

/**************** tridiagonal solver **************************/

void trid(REAL* a, REAL * b, REAL * c, REAL* d, int NX, int inc)
{
  int   i, ind=0;
  REAL aa, bb, cc, dd;

  bb     = 1.0f/b[ind];
  dd     = bb*d[ind];
  b[ind] = bb;
  d[ind] = dd;

  for (i=1; i<NX; i++) {
    cc     = bb*c[ind];
    c[ind] = cc;
    ind    = ind + inc;
    aa     = a[ind];
    bb     = b[ind] - aa*cc;
    dd     = d[ind] - aa*dd;							    						
    bb     = 1.0f/bb;
    dd     = bb*dd;
    d[ind] = dd;
    b[ind] = bb;
  }

  d[ind] = dd;

  for (i=NX-2; i>=0; i--) {
    ind    = ind - inc;
    dd     = d[ind] - c[ind]*dd;
    d[ind] = dd;
  }
}

// solver for case 2 (re-using factored matrices)

void trid1(REAL* a, REAL * b, REAL * c, REAL* d, int NX, int inc)
{
  int   i, ind=0;
  REAL aa, bb, dd;

  bb     = b[ind];
  dd     = bb*d[ind];
  d[ind] = dd;

  for (i=1; i<NX; i++) {
    ind    = ind + inc;
    aa     = a[ind];
    bb     = b[ind];
    dd     = bb*(d[ind]-aa*dd);
    d[ind] = dd;
  }

  d[ind] = dd;

// reverse pass

  for (i=NX-2; i>=0; i--) {
    ind = ind - inc;
    dd  = d[ind] - c[ind]*dd;
    d[ind] = dd;
  }
}


void adi3(int option, int J, int K, int L,
          REAL A0[], REAL A1[], REAL A2[],
          REAL V[], REAL R[], REAL R2[])
{
  int i, I;
  int j, k, l, ind;
  I = J*K*L;   
  switch (option) {

    /* standard ADI without Craig-Sneyd correction */

    case 0:
      for (l=0; l<L; l++) {
	  for (k=0; k<K; k++) {
	    ind = k*J + l*J*K;
	    trid(&A0[ind],&A0[ind+I],&A0[ind+2*I],&R[ind],J,1);
	   }
	}
	for (l=0; l<L; l++) {
	  for (j=0; j<J; j++) {
	   ind = j + l*J*K;
	   trid(&A1[ind],&A1[ind+I],&A1[ind+2*I],&R[ind],K,J);
	  }
	}
	for (k=0; k<K; k++) {
	  for (j=0; j<J; j++) {
	   ind = j + k*J;
	   trid(&A2[ind],&A2[ind+I],&A2[ind+2*I],&R[ind],L,J*K);
	  }
	} 
      for (i=0; i<I; i++) V[i] += R[i] ;
      break;

    /* version re-using factored matrices */

    case 2:
      for (l=0; l<L; l++) {
        for (k=0; k<K; k++) {
          ind = k*J + l*J*K;
          trid1(&A0[ind],&A0[ind+I],&A0[ind+2*I],&R[ind],J,1);
        }
      }

      for (l=0; l<L; l++) {
        for (j=0; j<J; j++) {
          ind = j + l*J*K;
          trid1(&A1[ind],&A1[ind+I],&A1[ind+2*I],&R[ind],K,J);
        }
      }

      for (k=0; k<K; k++) {
        for (j=0; j<J; j++) {
          ind = j + k*J;
          trid1(&A2[ind],&A2[ind+I],&A2[ind+2*I],&R[ind],L,J*K);
        }
      }

      for (i=0; i<I; i++) V[i] += R[i];
      break;

  }
}
