/*
* Copyright (c) 2008, Mike Giles
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * The name of Mike Giles may not be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY Mike Giles ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Mike Giles BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* GPU version of generic solver for 3D PDE's */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <myvar.h>  // defines whether REAL = "float" or "double"

////////////////////////////////////////////////////////////////////////
// define kernel block size for RHS calculation
////////////////////////////////////////////////////////////////////////

#define BLOCK_X 32
#define BLOCK_Y 4


/* top level driver for generic 3-factor PDE with abritrary drift and volatility */

void generic3_GPU(REAL Tmax, int M, REAL xmin[], REAL xmax[], int I, int J, int K,
                  REAL V[], char option[], REAL model_consts[], int model_consts_dim)
{
  int    m, pitch, IJK;
  REAL   dt, dx[3];
  REAL   *d_V, *d_A0, *d_A1, *d_A2, *d_R, *d_R2;
  size_t pitch_bytes;

/* allocate device arrays */

  CUDA_SAFE_CALL( cudaMallocPitch((void **)&d_V,  &pitch_bytes, sizeof(REAL)*I,3*J*K) );

  CUDA_SAFE_CALL( cudaMallocPitch((void **)&d_A0, &pitch_bytes, sizeof(REAL)*I,3*J*K) );
  CUDA_SAFE_CALL( cudaMallocPitch((void **)&d_A1, &pitch_bytes, sizeof(REAL)*I,3*J*K) );
  CUDA_SAFE_CALL( cudaMallocPitch((void **)&d_A2, &pitch_bytes, sizeof(REAL)*I,3*J*K) );
  CUDA_SAFE_CALL( cudaMallocPitch((void **)&d_R,  &pitch_bytes, sizeof(REAL)*I,  J*K) ); 
  CUDA_SAFE_CALL( cudaMallocPitch((void **)&d_R2, &pitch_bytes, sizeof(REAL)*I,  J*K) ); 
  
  pitch = pitch_bytes/sizeof(REAL);
  IJK   = pitch*J*K;

/* copy data to device */

  CUDA_SAFE_CALL(cudaMemcpy2D(d_V,pitch_bytes,V,sizeof(REAL)*I,
                              sizeof(REAL)*I,J*K,cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaThreadSynchronize());

/* copy constants to device */

  cudaMemcpyToSymbol(d_consts,model_consts,sizeof(REAL)*model_consts_dim);
 
/* grid spacing */

  dx[0]  = (xmax[0]-xmin[0]) / (I-1);
  dx[1]  = (xmax[1]-xmin[1]) / (J-1);
  dx[2]  = (xmax[2]-xmin[2]) / (K-1);

  dt = Tmax / M;

/* define two CUDA block sizes */

  dim3 dimGrid1(1+(I-1)/BLOCK_X, 1+(J-1)/BLOCK_Y);
  dim3 dimBlock1(BLOCK_X,BLOCK_Y);

  dim3 dimGrid2(1+(I-1)/16, 1+(J-1)/4);
  dim3 dimBlock2(16,4);

/* main Crank-Nicholson time-marching loop */
  
  for (m=0; m<M; m++) {

/* take standard timestep */

    if (strcmp(option,"CN1")==0){
      GPU_CN<<<dimGrid1,dimBlock1>>>(xmin[0],xmin[1],xmin[2],dx[0],dx[1],dx[2],dt,
                                     0, I, J, K, pitch, d_A0,d_A1,d_A2,d_R,d_V);
      CUDA_SAFE_CALL( cudaThreadSynchronize() );
      CUT_CHECK_ERROR("GPU_CN execution failed\n");

      GPU_adi_tridxy<<<dimGrid2,dimBlock2>>>(100,J,I,K,pitch,1,    d_A0,&d_A0[IJK],&d_A0[2*IJK],d_R);
      CUDA_SAFE_CALL( cudaThreadSynchronize() );
      CUT_CHECK_ERROR("GPU_adi_x execution failed\n");

      GPU_adi_tridxy<<<dimGrid2,dimBlock2>>>(200,I,J,K,pitch,pitch,d_A1,&d_A1[IJK],&d_A1[2*IJK],d_R);
      CUDA_SAFE_CALL( cudaThreadSynchronize() );
      CUT_CHECK_ERROR("GPU_adi_y execution failed\n");

      GPU_adi_tridz<<<dimGrid2,dimBlock2>>>(   0,I,J,K,pitch,  d_V,d_A2,&d_A2[IJK],&d_A2[2*IJK],d_R);
      CUDA_SAFE_CALL( cudaThreadSynchronize() );
      CUT_CHECK_ERROR("GPU_adi_z execution failed\n");        
    }       

/* Craig-Sneyd timestep */

    else if ( (strcmp(option,"CN2")==0)){

    }
  }

  /* copy data from device */

  CUDA_SAFE_CALL(cudaMemcpy2D(V,sizeof(REAL)*I,d_V,pitch_bytes,
                 sizeof(REAL)*I,J*K,cudaMemcpyDeviceToHost));
  CUDA_SAFE_CALL(cudaThreadSynchronize());

  /* deallocate arrays */

  CUDA_SAFE_CALL( cudaFree(d_V)  );

  CUDA_SAFE_CALL( cudaFree(d_A0) );
  CUDA_SAFE_CALL( cudaFree(d_A1) );
  CUDA_SAFE_CALL( cudaFree(d_A2) );
  CUDA_SAFE_CALL( cudaFree(d_R) );
  CUDA_SAFE_CALL( cudaFree(d_R2) );
}


