/*
* Copyright (c) 2008, Mike Giles
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * The name of Mike Giles may not be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY Mike Giles ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Mike Giles BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

//
// Notes:
//
// 1) strategy: one thread per node in the 2D block;
//    after initialisation it marches in the k-direction
//    working with 3 planes of data at a time
//
// 2) each thread also loads in data for at most one halo node;
//    assumes the number of halo nodes is not more than the
//    number of interior nodes
//
// 3) corner halo nodes are included because they are needed
//    for cross-derivatives


// definition to use efficient __mul24 intrinsic

#include <myvar.h>

#define INDEX(i,j,j_off)  (i +__mul24(j,j_off))

////////////////////////////////////////////////////////////////////////
// define kernel block size for RHS calculation
////////////////////////////////////////////////////////////////////////

#define BLOCK_X 32
#define BLOCK_Y 4


__constant__  REAL d_consts[100];



//
// This routine calculates the r.h.s. and the tridiagonal matrices
//

__global__ void GPU_CN(REAL xmin0,REAL xmin1,REAL xmin2,REAL dx0,REAL dx1,REAL dx2, REAL dt,
                       int option, int I, int J, int K, int pitch,
                       REAL A0[],REAL A1[],REAL A2[],REAL R[],REAL d_u[])
{
  int   indg, indg_h, indg0;
  int   i, j, k, i0,i1,i2, ind, ind_h, halo, active, IJK = pitch*J*K;
  REAL  xmin[3],dx[3],dxi[3],x[3],u[3],src[3],sig[3],beta[3];
  REAL  a0m,a0o,a0p,b0, a1m,a1o,a1p,b1, a2m,a2o,a2p,b2;

  xmin[0] = xmin0;
  xmin[1] = xmin1;
  xmin[2] = xmin2;

  dx[0] = dx0;
  dx[1] = dx1;
  dx[2] = dx2;

  dxi[0] = 1.0/dx0;
  dxi[1] = 1.0/dx1;
  dxi[2] = 1.0/dx2;

  // define local array offsets
 
#define IOFF  1
#define JOFF (BLOCK_X+2)
#define KOFF (BLOCK_X+2)*(BLOCK_Y+2)

   __shared__ REAL u1[3*KOFF];

  //
  // first set up indices for halos
  //

  k    = threadIdx.x + threadIdx.y*BLOCK_X;
  halo = k < 2*(BLOCK_X+BLOCK_Y+2);

  if (halo) {
    if (threadIdx.y<2) {               // y-halos (coalesced)
      i = threadIdx.x;
      j = threadIdx.y*(BLOCK_Y+1) - 1;
    }
    else {                             // x-halos (not coalesced)
      i = (k%2)*(BLOCK_X+1) - 1;
      j =  k/2 - BLOCK_X - 1;
    }

    ind_h  = INDEX(i+1,j+1,JOFF) + KOFF;
    u1[ind_h-KOFF]=0;u1[ind_h]=0;u1[ind_h+KOFF]=0;

    i      = INDEX(i,blockIdx.x,BLOCK_X);   // global indices
    j      = INDEX(j,blockIdx.y,BLOCK_Y);
    indg_h = INDEX(i,j,pitch);

    halo   =  (i>=0) && (i<I) && (j>=0) && (j<J);
  }

  //
  // then set up indices for main block
  //

  i    = threadIdx.x;
  j    = threadIdx.y;
  ind  = INDEX(i+1,j+1,JOFF) + KOFF;
  u1[ind-KOFF]=0;u1[ind]=0;u1[ind+KOFF]=0;

  i    = INDEX(i,blockIdx.x,BLOCK_X);     // global indices
  j    = INDEX(j,blockIdx.y,BLOCK_Y);
  indg = INDEX(i,j,pitch);

  active = (i<I) && (j<J);

  //
  // read initial plane of u1 array
  //

  if (active) u1[ind+KOFF] = d_u[indg];
  if (halo) u1[ind_h+KOFF] = d_u[indg_h];

  //
  // loop over k-planes
  //

  for (k=0; k<K; k++) {

    // move two planes down and read in new plane k+1

    if (active) {
      indg0 = indg;
      indg  = INDEX(indg,J,pitch);
      u1[ind-KOFF] = u1[ind];
      u1[ind]      = u1[ind+KOFF];
      if (k<K-1)
        u1[ind+KOFF] = d_u[indg];
    }

    if (halo) {
      indg_h = INDEX(indg_h,J,pitch);
      u1[ind_h-KOFF] = u1[ind_h];
      u1[ind_h]      = u1[ind_h+KOFF];
      if (k<K-1)
        u1[ind_h+KOFF] = d_u[indg_h];
    }

    __syncthreads();

    //
    // calculate r.h.s. and set a,b,c, coefficients
    //

    if (active) {
      params(i,j,k,I,J,K,dt,x,dx,xmin,dxi, src,u,sig,beta, d_consts);

      a0m =      sig[0] - u[0];
      a0o = -2.0*sig[0] - src[0];
      a0p =      sig[0] + u[0];

      a1m =      sig[1] - u[1];
      a1o = -2.0*sig[1] - src[1];
      a1p =      sig[1] + u[1];

      a2m =      sig[2] - u[2];
      a2o = -2.0*sig[2] - src[2];
      a2p =      sig[2] + u[2];

      i0 = (beta[0]<0? -1:1);
      i1 = (beta[1]<0? -1:1);
      i2 = (beta[2]<0? -1:1);

      b0 = fabs(beta[0]);
      b1 = fabs(beta[1]);
      b2 = fabs(beta[2]);

      R[indg0] = (a0m-b1-b2)*u1[ind-1   ] + (a0p-b1-b2)*u1[ind+1   ] 
               + (a1m-b2-b0)*u1[ind-JOFF] + (a1p-b2-b0)*u1[ind+JOFF]
               + (a2m-b0-b1)*u1[ind-KOFF] + (a2p-b0-b1)*u1[ind+KOFF]
               + (a0o+a1o+a2o+2.0*(b0+b1+b2))*u1[ind]
               + b2*(u1[ind-1   -i2*JOFF] + u1[ind+1   +i2*JOFF])
               + b1*(u1[ind-1   -i1*KOFF] + u1[ind+1   +i1*KOFF])
               + b0*(u1[ind-JOFF-i0*KOFF] + u1[ind+JOFF+i0*KOFF]);

      A0[indg0      ] =     - 0.5*a0m;
      A0[indg0+IJK  ] = 1.0 - 0.5*a0o;
      A0[indg0+2*IJK] =     - 0.5*a0p;

      A1[indg0      ] =     - 0.5*a1m;
      A1[indg0+IJK  ] = 1.0 - 0.5*a1o;
      A1[indg0+2*IJK] =     - 0.5*a1p;

      A2[indg0      ] =     - 0.5*a2m;
      A2[indg0+IJK  ] = 1.0 - 0.5*a2o;
      A2[indg0+2*IJK] =     - 0.5*a2p;
    }

    __syncthreads();

  }
}

//
// tri-diagonal solve in x- and y-directions
//

__global__ void GPU_adi_tridxy(int dim,int NX, int NY, int NZ, int pitch, int inc,
                               REAL *d_a, REAL *d_b, REAL *d_c, REAL *d_d)
{
  int   i, j, k, indg;
  REAL aa, bb, cc, dd;

  //
  // set up indices for main block
  //

  i    = INDEX(threadIdx.x,blockIdx.x,blockDim.x);  // global indices
  k    = INDEX(threadIdx.y,blockIdx.y,blockDim.y);
  if (dim==100)
  indg = INDEX(0,pitch,INDEX(i,k,NX));
  else if (dim==200)
  indg = INDEX(i,k,INDEX(0,pitch,NY));
  

  if ( (i<NX) && (k<NZ) ) {

  // forward pass

    bb   = 1.0f/d_b[indg];
    dd   = bb*d_d[indg];
    d_b[indg]=bb;
    d_d[indg] = dd;

    for (j=1; j<NY; j++) {
      cc        = bb*d_c[indg];
      d_c[indg] = cc;
      indg +=inc;
      aa   = d_a[indg];
      bb   = d_b[indg] - aa*cc;
      dd   = d_d[indg] - aa*dd;
      bb   = 1.0f/bb;
      dd   = bb*dd;
      d_d[indg] = dd;
      d_b[indg] = bb;
    }

  // reverse pass

    d_d[indg] = dd;

    for (j=NY-2; j>=0; j--) {
      indg = indg - inc;
      dd = d_d[indg] - d_c[indg]*dd;
      d_d[indg] = dd;
    }
  }
}

//
// tri-diagonal solve in z-direction 
//

__global__ void GPU_adi_tridz(int option,int NX, int NY, int NZ, int pitch, REAL *d_u,
                          REAL *d_a, REAL *d_b, REAL *d_c, REAL *d_d)
{
  int   i, j, k, indg, off;
  REAL aa, bb, cc, dd;

  //
  // set up indices for main block
  //

  i    = INDEX(threadIdx.x,blockIdx.x,blockDim.x);  // global indices
  j    = INDEX(threadIdx.y,blockIdx.y,blockDim.y);
  indg = INDEX(i,j,pitch);
  off  = INDEX(0,pitch,NY);

  if ( (i<NX) && (j<NY) ) {

  // forward pass

    bb   = 1.0f/d_b[indg];
    dd   = bb*d_d[indg];
    d_b[indg]=bb;
    d_d[indg] = dd;

    for (k=1; k<NZ; k++) {
      cc        = bb*d_c[indg];
      d_c[indg] = cc;
      indg +=off;
      aa   = d_a[indg];
      bb   = d_b[indg] - aa*cc;
      dd   = d_d[indg] - aa*dd;
      bb   = 1.0f/bb;
      dd   = bb*dd;
      d_d[indg] = dd;
      d_b[indg] = bb;
    }

  // reverse pass

    if (option!=1) { 
      d_d[indg] = dd;
      d_u[indg] = d_u[indg]+dd;     
    }
    else
      d_d[indg] = dd;


    for (k=NZ-2; k>=0; k--) {
      indg = indg - off;
      dd = d_d[indg] - d_c[indg]*dd;
      if (option!=1) { 
        d_d[indg] = dd;
        d_u[indg] = d_u[indg]+dd;
      }
      else
        d_d[indg] = dd;
    }
  }
}


__global__ void GPU_adi_trid1(int dim,int NX, int NY, int NZ, int pitch,int inc, REAL *d_u,
                          REAL *d_a, REAL *d_b, REAL *d_c, REAL *d_d)
{
  int   i, j, k, indg;
  REAL aa, bb, dd;

  //
  // set up indices for main block
  //

  i    = INDEX(threadIdx.x,blockIdx.x,blockDim.x);  // global indices
  k    = INDEX(threadIdx.y,blockIdx.y,blockDim.y);
  if (dim==100)
  indg = INDEX(0,pitch,INDEX(i,k,NX));
  else if (dim==200)
  indg = INDEX(i,k,INDEX(0,pitch,NY));
  if (dim==300)
  {
  indg = INDEX(i,k,pitch);
  inc=  INDEX(0,pitch,NZ);
  }

  if ( (i<NX) && (k<NZ) ) {

  // forward pass

    bb   = d_b[indg];
    dd   = bb*d_d[indg];
    d_d[indg] = dd;

    for (j=1; j<NY; j++) {
      indg = indg + inc;
      aa   = d_a[indg];
      bb   = d_b[indg];
      dd   = bb*(d_d[indg]-aa*dd);
      d_d[indg] = dd;
    }

  // reverse pass

    if (dim==300)
      {
        d_d[indg] = dd;
        d_u[indg] = d_u[indg]+dd; 
      }
      else
        d_d[indg] = dd;


    for (j=NY-2; j>=0; j--) {
      indg = indg - inc;
      dd = d_d[indg] - d_c[indg]*dd;
      if (dim==300)
      {
        d_d[indg] = dd;
        d_u[indg] = d_u[indg]+dd; 
      }
      else
        d_d[indg] = dd;
    }
  }
}
