/*
* Copyright (c) 2008, Mike Giles
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * The name of Mike Giles may not be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY Mike Giles ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Mike Giles BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* Test of generic 3D PDE solver for GPU execution*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <cutil.h>
#include <string.h>

#include <myvar.h>   // defines whether REAL = "float" or "double"

////////////////////////////////////////////////////////////////////////
// specification of the model's drift, volatility, source terms
////////////////////////////////////////////////////////////////////////

__host__ __device__ void model(REAL *x,REAL *src, REAL *u, REAL *sigma, REAL *beta, REAL *consts)
{
  sigma[0] = consts[3];
  sigma[1] = consts[4];
  sigma[2] = consts[5];

  beta[0] = consts[6];
  beta[1] = consts[7];
  beta[2] = consts[8];

  src[0] = 0.0;
  src[1] = 0.0;
  src[2] = 0.0;

  u[0] = -consts[0]*x[0];
  u[1] = -consts[1]*x[1];
  u[2] = -consts[2]*x[2];
}

////////////////////////////////////////////////////////////////////////
// include generic routines
////////////////////////////////////////////////////////////////////////

#include <generic_3d_params.cu>          // routine used on both GPU and CPU

#include <generic_3d_GPU_kernels.cu>  // kernel routines that execute on GPU
#include <generic_3d_GPU.cu>          // host routine which lanuches kernels

#include <generic_3d_CPU.cu>          // CPU version for comparison


int main(int argc, char *argv[])
{  
  int  I, J, K, N, Ip, Jp, Kp, IJ, mid, pass, Ns[5], i, j, k, opt;
  REAL Tmax, xmin[3], xmax[3], dx[3], x[3], *V, *V1;
  REAL lam[3], sigma[3], beta[3], model_consts[9];
  char mropt[8];

  unsigned int Timer;

/* initialise device */

  CUT_DEVICE_INIT();
  CUT_SAFE_CALL( cutCreateTimer(&Timer) );

/* modelling constants */

  lam[0]   = 0.05;
  lam[1]   = 0.2;
  lam[2]   = 0.5;

  sigma[0] = 0.025;
  sigma[1] = 0.015;
  sigma[2] = 0.01;

  beta[0] = 0.5;
  beta[1] = 0.3;
  beta[2] = 0.1;

  for (i=0; i<3; i++) {
    model_consts[i]   = lam[i];
    model_consts[i+3] = sigma[i];
    model_consts[i+6] = beta[i];
  }

/* computational grid */

  Tmax = 5.0;  

  xmax[0] = 5.0*sigma[0]/sqrt(lam[0]);
  xmax[1] = 5.0*sigma[1]/sqrt(lam[1]);
  xmax[2] = 5.0*sigma[2]/sqrt(lam[2]);
  xmin[0] = -xmax[0];
  xmin[1] = -xmax[1];
  xmin[2] = -xmax[2];

/* main execution loop */


  for (opt=0; opt<1; opt++){
    // printf("\nopt = %d\n",opt);

    if (opt==0)
      strcpy(mropt,"CN1");
    else if (opt==1)
      strcpy(mropt,"CN2");

    printf("\nopt: %s\n",mropt);

    Ns[0] = 20;
    Ns[1] = 40;
    Ns[2] = 80;

    for (pass=0; pass<3; pass++) {
      N = Ns[pass];
      I = N;
      J = N;
      K = N;

      Ip  = I+1;
      Jp  = J+1;
      Kp  = K+1;
      IJ  = Ip*Jp;
      mid = I/2 + Ip*J/2 + IJ*K/2;

      dx[0]=(xmax[0]-xmin[0]) / I;
      dx[1]=(xmax[1]-xmin[1]) / J;
      dx[2]=(xmax[2]-xmin[2]) / K;

/* allocate host memory with extra zeros on either side */

      V  = (REAL*)calloc(2*Ip*Jp*Kp+2*(Ip+1)*(Jp+1),sizeof(REAL))+(Ip+1)*(Jp+1);
      V1 = (REAL*)calloc(2*Ip*Jp*Kp+2*(Ip+1)*(Jp+1),sizeof(REAL))+(Ip+1)*(Jp+1);

/* define payoff vector */

      for (k=0; k<Kp; k++){
        for (j=0; j<Jp; j++) {
          for (i=0; i<Ip; i++) {
            x[0] = xmin[0] + i*dx[0];
            x[1] = xmin[1] + j*dx[1];
            x[2] = xmin[2] + k*dx[2];
            V[i+j*Ip+k*Ip*Jp] = x[0]*x[1]*1000;
            V1[i+j*Ip+k*Ip*Jp] = V[i+j*Ip+k*Ip*Jp];
          }
        }
      }
   
/* numerical solution */

      if (opt==0) {                    // standard solution
        CUT_SAFE_CALL(cutStartTimer(Timer));
        generic3_CPU(Tmax, N, xmin, xmax ,Ip, Jp, Kp, V, mropt, model_consts,9);
        CUT_SAFE_CALL(cutStopTimer(Timer));
        printf("CPU............N, V, time = %3d, %14.9g  %f (ms)\n",
                N, V[mid], cutGetTimerValue(Timer));
        CUT_SAFE_CALL(cutResetTimer(Timer));

        CUT_SAFE_CALL(cutStartTimer(Timer));
        generic3_GPU(Tmax, N, xmin, xmax ,Ip, Jp, Kp, V1, mropt, model_consts,9);
        CUT_SAFE_CALL(cutStopTimer(Timer));
        printf("GPU............N, V, time = %3d, %14.9g  %f (ms)\n",
                N, V1[mid], cutGetTimerValue(Timer));
        CUT_SAFE_CALL(cutResetTimer(Timer));
      }

      free(V  - (Ip+1)*(Jp+1));
      free(V1 - (Ip+1)*(Jp+1));
    }
  }

  CUT_SAFE_CALL(cutDeleteTimer(Timer));
  // CUT_EXIT(argc,argv);
}
